/* 
file : metropolis.hpp 
Defenition of the metropolis functor. This class is called as a 
function, operating on the lattice graph using the hamiltonian 
functor for a given number of times (sweeps).

Spinner v.2, a Monte Carlo program for the simulation of 'classical'
magnetic systems.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef METROPOLIS
#define METROPOLIS

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/lagged_fibonacci.hpp> 
#include <boost/random/variate_generator.hpp>
#include <boost/random/uniform_on_sphere.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/uniform_real.hpp>
//#include <random/uniform.h>
//#include <random/discrete-uniform.h>

#include "flippers.hpp"
#include "math_utility.hpp"

struct metropolis_Ising
{
public:
  metropolis_Ising(po::variables_map&,map_of_fifos&,lattice_graph&);
  ~metropolis_Ising();
  template<class hamiltonian,class flipper>
  void operator()(double, uint,lattice_graph&,map_of_fifos&,
		  hamiltonian&,flipper&);
protected :
  uint i,j;
  uint Nspins;
  double E,Enew,beta;
  long double P;
  euclid_vector M,R;
  po::variables_map vm;
  site_iterator s,first_site,last_site;
  edge_iterator edge_iter, first_edge, last_edge, next_edge;
  o_edge_iterator out_i, out_end;
  graph_traits<lattice_graph>::vertex_descriptor site;
  graph_traits<lattice_graph>::edge_descriptor edge;

  //random number generators apart from dimension dependent uniform on 
  //sphere.
  mt19937 mersenne_gen; // to pick a spin
  lagged_fibonacci607 fib_gen,fib_gen2;  
  uniform_int<int> uni_int_dist;
  uniform_real<long double> uni_01_dist;
  variate_generator<mt19937&, uniform_int<int> > uni_int;
  variate_generator<lagged_fibonacci607&, uniform_real<long double> > uni_01;

};


struct metropolis : public metropolis_Ising
{
public:
  metropolis(po::variables_map&,map_of_fifos&,lattice_graph&);
  template<class hamiltonian,class flipper>
  void operator()(double, uint, lattice_graph&, map_of_fifos&,
		  hamiltonian&,flipper&);
protected :
  uniform_on_sphere<double,euclid_vector> uni_sphere;
  variate_generator<lagged_fibonacci607&, 
		    uniform_on_sphere<double,euclid_vector> > fib_sphere; 
  uint n,flips;
  euclid_vector old_spin;
  double phi,phi_max,phi_up_lim,phi_lo_lim;
  std::vector<std::string> M_i;
  ublas::zero_vector<double> zero;
};


struct metropolis_heisenberg : public metropolis_Ising
{
public:
  metropolis_heisenberg(po::variables_map&,map_of_fifos&,lattice_graph&);
  template<class hamiltonian,class flipper>
  void operator()(double, uint,lattice_graph&,map_of_fifos&,
		  hamiltonian&,flipper&);
protected:
  uniform_on_sphere<double,euclid_vector> uni_sphere;
  variate_generator<lagged_fibonacci607&, 
		    uniform_on_sphere<double,euclid_vector> > fib_sphere; 

  uint flips;
  euclid_vector old_spin;
  double phi,phi_max;
  ublas::zero_vector<double> zero;
};


struct Qmetro : public metropolis_heisenberg
{
public:
  Qmetro(po::variables_map&,map_of_fifos&,lattice_graph&);

  template<class hamiltonian,class flipper>
  void operator()(double, uint, lattice_graph&, map_of_fifos&,
		  hamiltonian&,flipper&);
protected:
  uint Nbonds;
  euclid_vector old_S2;
  mt19937 mersenne_gen2; // to pick a spin
  uniform_int<int> allBonds;
  variate_generator<mt19937&, uniform_int<int> > pickBond;
};

////////////////////////////////////////////////////////////////////////////
//////////////////////////// IMPLEMENTATIONS ///////////////////////////////
////////////////////////////////////////////////////////////////////////////


metropolis_Ising::metropolis_Ising
(po::variables_map& VM,map_of_fifos& dm,lattice_graph& LG)  
  : vm(VM), Nspins(num_vertices(LG)),
    uni_int_dist(0,num_vertices(LG)-1), 
    uni_int(mersenne_gen, uni_int_dist),
    uni_01_dist(0,1), 
    uni_01(fib_gen2,uni_01_dist)
{
  tie(first_site,last_site)=vertices(LG);
  tie(first_edge,last_edge)=edges(LG);

  std::ifstream state_file;
  if (vm.count("random-state")) 
    state_file.open(vm["random-state"].as<std::string>().c_str(),std::ios::in);
  if (state_file) state_file>>mersenne_gen>>fib_gen>>fib_gen2;
  else {
    mersenne_gen.seed(static_cast<uint>(time(0)));    
    fib_gen.seed(static_cast<uint>(time(0))+10);
    fib_gen2.seed(static_cast<uint>(time(0))+100);
    tie(first_site,last_site)=vertices(LG);
  }
}


metropolis_Ising::~metropolis_Ising()
{
  if (vm.count("random-state")) {
    std::ofstream state_file(vm["random-state"].as<std::string>().c_str(),std::ios::out);
    state_file<<mersenne_gen<<fib_gen<<fib_gen2;
  }
}


template<class hamiltonian,class flipper>
void metropolis_Ising::operator()
  (double temp, uint sweeps,lattice_graph& LG,map_of_fifos&
   dm,hamiltonian& H,flipper& flip)
  {
    beta = 1 / temp;
    for (i=0; i < sweeps; i++){
      for (j=0; j < Nspins; j++){
	// pick a random spin
	// t is the iterator pointing to the randomly chosen spin.
	s=first_site+uni_int(); 
	// calc energy
	E = H(out_edges(*s,LG).first,out_edges(*s,LG).second,s,s+1,LG);
	// flip spin
	LG[*s].spin = -LG[*s].spin;
	// calc new energy
	Enew = H(out_edges(*s,LG).first,out_edges(*s,LG).second,s,s+1,LG);
	// accept or reject.
	P = exp ((E - Enew)*beta);
	if (P < uni_01 ()) 
	  LG[*s].spin=-LG[*s].spin; //back to old pos
      }
      // As a convention, keep variables in dm which need individual treatment
      // in capital, and the ones which can be treated as m_x,m_y,m_z etc. in lower case.
      // get total energy
      dm["E"].push_back(H(first_edge,last_edge,first_site,last_site,LG));
      // get total magnetisation
      M[0]=0;
      for (s=first_site; s!=last_site; s++) M+=LG[*s].spin;
      dm["Mabs"].push_back(sqrt(M[0]*M[0]));
      dm["Mabs4"].push_back(M[0]*M[0]*M[0]*M[0]);
      dm["m_x"].push_back(M[0]);

      // Add your own observables below if needed:
      // dm["your own variable"].push_back(Value of your own variable);

    }
  }


metropolis::metropolis
(po::variables_map& VM,map_of_fifos& dm,lattice_graph& LG)  
  : metropolis_Ising (VM,dm,LG),
    uni_sphere(VM["spin-dimension"].as<int>()), 
    fib_sphere(fib_gen, uni_sphere),
    zero(VM["spin-dimension"].as<int>()),
    old_spin(VM["spin-dimension"].as<int>())
{
  M_i.push_back("m_x"); M_i.push_back("m_y"); M_i.push_back("m_z");
  phi_lo_lim=0.02; phi_up_lim=M_PI;
  phi_max = M_PI;
  flips = 0;
}


template<class hamiltonian, class flipper>
void metropolis::operator()
  (double temp, uint sweeps,lattice_graph& LG,map_of_fifos& dm,hamiltonian& H,flipper& flip)
{
  beta = 1 / temp;
  for (i=0; i < sweeps; i++){
    for (j=0; j < Nspins; j++){
	// pick a random spin
	// t is the iterator pointing to the randomly chosen spin.
      s=first_site+uni_int(); 
      // calc energy
      E = H(out_edges(*s,LG).first,out_edges(*s,LG).second,s,s+1,LG);
      // flip spin to random direction
      old_spin = LG[*s].spin;
      R = fib_sphere();
      flip(s,LG, R, phi_max);
      // calc new energy
      Enew = H(out_edges(*s,LG).first,out_edges(*s,LG).second,s,s+1,LG);
      // accept or reject.
      P = exp ((E - Enew)*beta);
      if (P < uni_01 ()) 
	LG[*s].spin=old_spin;
      else // count the number of accepts. 
	flips++;
    }
    // As a convention, keep variables in dm which need individual treatment
    // in capital, and the ones which can be treated as m_x,m_y,m_z etc. in lower case.

    // get total energy
    dm["E"].push_back(H(first_edge,last_edge,first_site,last_site,LG));
    // get total magnetisation
    M=zero; n=0; 
    for (s=first_site; s!=last_site; s++) M+=LG[*s].spin;
    dm["Mabs"].push_back(sqrt(ublas::inner_prod(M,M)));
    dm["Mabs4"].push_back(pow(ublas::inner_prod(M,M),2));
    for (n=0;n<M.size();n++) dm[M_i[n]].push_back(M[n]);

    // Add your own observables below if needed:
    // dm["your own variable"].push_back(Value of your own variable);
    
    // adjust max rotation angle of spin to get
    // accept rate close of 50%.
    phi_max=phi_max*2*flips/Nspins;
    if (phi_max<0.02) phi_max=0.02;
    else if (phi_max>M_PI) phi_max=M_PI;
    flips = 0;
  }
}


metropolis_heisenberg::metropolis_heisenberg
(po::variables_map& VM,map_of_fifos& dm,lattice_graph& LG)  
  : metropolis_Ising (VM,dm,LG),
    uni_sphere(3),    
    fib_sphere(fib_gen, uni_sphere),
    zero(3),
    old_spin(3)
{ 
  std::cout<<"SIMUTATION TYPE: Heisenberg\n";
  phi_max = M_PI;
}


template<class hamiltonian,class flipper> //flipper is a dummy here!
void metropolis_heisenberg::operator()
  (double temp, uint sweeps,lattice_graph& LG,map_of_fifos&
   dm,hamiltonian& H,flipper& flip)
{
  beta = 1 / temp;
  for (i=0; i < sweeps; i++){
    for (j=0; j < Nspins; j++){
      // pick a random spin
      // t is the iterator pointing to the randomly chosen spin.
      s=first_site+uni_int();
      // calc energy
      E = H(out_edges(*s,LG).first,out_edges(*s,LG).second,s,s+1,LG);
      // flip spin to random direction
      old_spin = LG[*s].spin;
      R = fib_sphere();
      phi=phi_max*(2*uni_01()-1);
      LG[*s].spin = cos(phi) * old_spin + 
	(ublas::inner_prod(R,old_spin)*(1-cos(phi))) * R + 
	sin(phi) * outer_product(R,old_spin); 
      // calc new energy
      Enew = H(out_edges(*s,LG).first,out_edges(*s,LG).second,s,s+1,LG);
      //std::cout<<"Eo = "<<E<<" En = "<<Enew<<std::endl;
      // accept or reject.
      P = exp ((E - Enew)*beta);
      if (P < uni_01 ()) 
	LG[*s].spin=old_spin;
      else // count the number of accepts. 
	flips++;
    }
    // get total energy
    dm["E"].push_back(H(first_edge,last_edge,first_site,last_site,LG));
    
    // get total magnetisation
    M=zero; for (s=first_site; s!=last_site; s++) M+=LG[*s].spin;
    dm["Mabs"].push_back(sqrt(ublas::inner_prod(M,M)));
    dm["Mabs4"].push_back(pow(ublas::inner_prod(M,M),2));
    // As a convention, keep variables in dm which need individual treatment
    // in capital, and the ones which can be treated as m_x,m_y,m_z etc. in lower case.
    dm["m_x"].push_back(M[0]); 
    dm["m_y"].push_back(M[1]);
    dm["m_z"].push_back(M[2]);
    
    // Add your own observables below if needed:
    // dm["your own variable"].push_back(Value of your own variable);
    
    // adjust max rotation angle of spin of get
    // accept rate close of 50%.
    phi_max=phi_max*2*flips/Nspins;
    if (phi_max<0.02) phi_max=0.02;
    else if (phi_max>M_PI) phi_max=M_PI;
    flips = 0;
  }
}


#endif 

