#ifndef METROPOLIS
#define METROPOLIS

#include <iostream>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/lagged_fibonacci.hpp> 
#include <boost/random/variate_generator.hpp>
#include <boost/random/uniform_on_sphere.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/uniform_real.hpp>
//#include <random/uniform.h>
//#include <random/discrete-uniform.h>

#include "math_utility.hpp"

struct metropolis_base
{
public:
  metropolis_base(po::variables_map&,map_of_fifos&,lattice_graph&);
  template<class hamiltonian>
  void operator()(double, uint,lattice_graph&,map_of_fifos&,hamiltonian&){}
protected :
  uint i,j;
  uint Nspins,flips;
  double E,Enew,phi,phi_max,beta;
  long double P;
  euclid_vector old_spin,M,R;
  po::variables_map vm;
  site_iterator s,first_site,last_site;
  edge_iterator edge_iter, last_edge;
  o_edge_iterator out_i, out_end;
  graph_traits<lattice_graph>::vertex_descriptor site;
  graph_traits<lattice_graph>::edge_descriptor edge;
};

struct metropolis_boost_rng : public metropolis_base
{
public:
  metropolis_boost_rng(po::variables_map&,map_of_fifos&,lattice_graph&);
  ~metropolis_boost_rng();
  template<class hamiltonian>
  void operator()(double, uint,lattice_graph&,map_of_fifos&,hamiltonian&);
protected :
  mt19937 mersenne_gen; // to pick a spin
  lagged_fibonacci607 fib_gen,fib_gen2;  
  //lagged_fibonacci4423 fig_gen2;
  uniform_int<int> uni_int_dist;
  uniform_real<long double> uni_01_dist;
  uniform_on_sphere<double,euclid_vector> uni_sphere;
  variate_generator<mt19937&, uniform_int<int> > uni_int;
  variate_generator<lagged_fibonacci607&, uniform_real<long double> > uni_01;
  variate_generator<lagged_fibonacci607&, 
		    uniform_on_sphere<double,euclid_vector> > fib_sphere; 
};

/*
struct metropolis_blitz_rng : public metropolis_base
{
public:
  metropolis_blitz_rng(po::variables_map&,map_of_fifos&,lattice_graph&);
  template<class hamiltonian>
  void operator()(double, uint,lattice_graph&,map_of_fifos&,hamiltonian&);
protected :
  ranlib::Uniform < double > uniPGen,uniSGen;
  ranlib::DiscreteUniform < unsigned int >discrUniGen;
};*/


metropolis_base::metropolis_base
(po::variables_map& VM,map_of_fifos& dm,lattice_graph& LG)  
  : vm(VM), Nspins(num_vertices(LG))
{
  dm["E"]=empty_fifo;
  dm["M_x"]=empty_fifo;
  dm["M_y"]=empty_fifo;
  dm["M_z"]=empty_fifo;
  dm["|M|"]=empty_fifo;
  tie(first_site,last_site)=vertices(LG);
  phi_max = M_PI;
  flips = 0;
}

metropolis_boost_rng::metropolis_boost_rng
(po::variables_map& VM,map_of_fifos& dm,lattice_graph& LG)  
  : metropolis_base (VM,dm,LG), 
    uni_sphere(3), 
    uni_int_dist(0,num_vertices(LG)-1), 
    uni_01_dist(0,1), 
    uni_01(fib_gen2,uni_01_dist), 
    fib_sphere(fib_gen, uni_sphere), 
    //fib_gen2(),
    uni_int(mersenne_gen, uni_int_dist) 
{
  std::cout<<"initializing random number generators..."<<std::flush;
  std::ifstream state_file;
  if (vm.count("random-state")) 
    state_file.open(vm["random-state"].as<std::string>().c_str(),std::ios::in);
  if (state_file) state_file>>mersenne_gen>>fib_gen>>fib_gen2;
  else {
    mersenne_gen.seed(static_cast<uint>(time(0)));    
    fib_gen.seed(static_cast<uint>(time(0)));
    fib_gen2.seed(static_cast<uint>(time(0)));
    tie(first_site,last_site)=vertices(LG);
  }
  std::cout<<" ok"<<std::endl;
}


metropolis_boost_rng::~metropolis_boost_rng()
{
  if (vm.count("random-state")) {
    std::ofstream state_file(vm["random-state"].as<std::string>().c_str(),std::ios::out);
    state_file<<mersenne_gen<<fib_gen<<fib_gen2;
  }
}

/*
metropolis_blitz_rng::metropolis_blitz_rng
(po::variables_map& VM,map_of_fifos& dm,lattice_graph& LG)  
  : metropolis_base (VM,dm,LG), 
    discrUniGen((num_vertices(LG)-1))
{
  uniPGen.seed ((unsigned int) time (0));
  discrUniGen.seed ((2 * (unsigned int) time (0)));
  uniSGen.seed((3*(unsigned int)time(0)));
}


template<class hamiltonian>
void metropolis_blitz_rng::operator()
  (double temp, uint sweeps,lattice_graph& LG,map_of_fifos& dm,hamiltonian& H)
  {
    for (i=0; i < sweeps; i++){
      flips = 0;
      for (j=0; j < Nspins; j++){
	//pick a random spin
	// t is the iterator pointing to the randomly chosen spin.
	t=first_site+discrUniGen.random(); 
	//calc energy
	E = H(out_edges(*t,LG).first,out_edges(*t,LG).second,LG);
	//flip to random direction
	old_spin = LG[*t].spin;
	do {
	  R[0]=2*uniSGen.random()-1;
	  R[1]=2*uniSGen.random()-1;
	  R[2]=2*uniSGen.random()-1;
	} while (ublas::inner_prod(R,R)>1);
	phi=phi_max*(2*uniSGen.random()-1);
	R=R/sqrt(ublas::inner_prod(R,R));
	LG[*t].spin = cos(phi)*old_spin + 
	  ublas::inner_prod(R,old_spin)*(1-cos(phi))*R + 
	  outer_product(R,old_spin)*sin(phi);
	//calc energy
	Enew = H(out_edges(*t,LG).first,out_edges(*t,LG).second,LG);
	//accept or reject.
	P = exp (-(Enew - E) / temp);
	if (P < uniPGen.random())
	  LG[*t].spin=old_spin;
	else
	  flips++;
      }
      //get total energy
      dm["E"].push_back(H(edges(LG).first,edges(LG).second,LG));

      //get total magnetisation
      M=zero; for (t=first_site; t!=last_site; t++) M+=LG[*t].spin;
      dm["M_x"].push_back(M[0]); 
      dm["M_y"].push_back(M[1]);
      dm["M_z"].push_back(M[2]);
      dm["|M|"].push_back(sqrt(ublas::inner_prod(M,M)));
      //std::cout<<"flips : "<<flips<<"\t"<<std::flush;
      phi_max=phi_max*2*flips/Nspins;
      if (phi_max<0.02) phi_max=0.02;
      if (phi_max>M_PI) phi_max=M_PI;
    }
  }
*/
  
template<class hamiltonian>
void metropolis_boost_rng::operator()
  (double temp, uint sweeps,lattice_graph& LG,map_of_fifos& dm,hamiltonian& H)
  {
    beta = 1 / temp;
    for (i=0; i < sweeps; i++){
      for (j=0; j < Nspins; j++){
	// pick a random spin
	// t is the iterator pointing to the randomly chosen spin.
	s=first_site+uni_int(); 
	//std::cout<<*t<<'\t';
	// calc energy
	E = H(out_edges(*s,LG).first,out_edges(*s,LG).second,LG);
	// flip spin to random direction
	old_spin = LG[*s].spin;
	R = fib_sphere();
	phi=phi_max*(2*uni_01()-1);
	LG[*s].spin = cos(phi) * old_spin + 
	  (ublas::inner_prod(R,old_spin)*(1-cos(phi))) * R + 
	  sin(phi) * outer_product(R,old_spin); 
	// calc new energy
	Enew = H(out_edges(*s,LG).first,out_edges(*s,LG).second,LG);
	// accept or reject.
	P = exp ((E - Enew)*beta);
	if (P < uni_01 ()) 
	  LG[*s].spin=old_spin;
	else // count the number of accepts. 
	  flips++;
      }
      //std::cout<<LG[*t].spin<<'\t';
      // get total energy
      dm["E"].push_back(H(edges(LG).first,edges(LG).second,LG));

      // get total magnetisation
      M=zero; for (s=first_site; s!=last_site; s++) M+=LG[*s].spin;
      dm["M_x"].push_back(M[0]); 
      dm["M_y"].push_back(M[1]);
      dm["M_z"].push_back(M[2]);
      dm["|M|"].push_back(sqrt(ublas::inner_prod(M,M)));

      // adjust max rotation angle of spin of get
      // accept rate close of 50%.
      phi_max=phi_max*2*flips/Nspins;
      if (phi_max<0.02) phi_max=0.02;
      else if (phi_max>M_PI) phi_max=M_PI;
      flips = 0;
    }
  }

#endif 
