/* 
file: options.cpp
Implements the command line interface. 

Spinner v.2, a Monte Carlo program for the simulation of 'classical'
magnetic systems.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "options.hpp"
#include "banners.hpp"

std::vector<double> parse_doubles(const std::vector<std::string>& values)
{
  double entry;
  std::vector<double> doubles;
  char_separator<char> sep(";|,) ");
  //TODO values is of type vector<string>, so must look over here.
  std::vector<std::string>::const_iterator value=values.begin();
  for (;value!=values.end();value++){
    tokenizer<char_separator<char> > atom_tokens(*value,sep);
    tokenizer<char_separator<char> >::iterator it;
    for(it=atom_tokens.begin(); it!=atom_tokens.end();++it) {
      entry=atof((*it).c_str()); 
      doubles.push_back(entry); 
    }
  }
  return doubles;
}

euclid_vector parse_vector(const int dim,const std::string& values)
{
  double entry;
  euclid_vector doubles(dim);
  char_separator<char> sep("(;|,) ");
  //TODO values is of type vector<string>, so must look over here.
  tokenizer<char_separator<char> > toks(values,sep);
  tokenizer<char_separator<char> >::iterator it;

  int n=0;
  for(it=toks.begin();it!=toks.end();it++){
    if (n > dim-1)
      std::cout<<"Warning : Not all components in '"<<values.c_str()<<"' are used.";
    else {
      entry=atof((*it).c_str()); 
      doubles[n]=entry; n++; 
    }
  }
  if (n < dim) 
    throw po::invalid_option_value("Vector entry has too few components.");
  return doubles;
}

euclid_matrix parse_matrix(const std::string& values)
{
  double entry;
  euclid_matrix doubles;
  char_separator<char> sep("(;|,)[] ");

  tokenizer<char_separator<char> > toks(values,sep);
  tokenizer<char_separator<char> >::iterator it;

  int n=0;
  for(it=toks.begin();it!=toks.end();it++){
    if (n > 8)
      throw po::invalid_option_value("Tensor variables must have 9 components.");
    entry=atof((*it).c_str());
    doubles(n/3,n%3)=entry; n++;
  }
  if (n < 8)
    throw po::invalid_option_value("Tensor variables must have 9 components.");
  return doubles;
}

po::variables_map parse_options(int ac, char* av[],std::string fn)
  // if fn is left empty then an input-file is expected from the command
  // line, otherwise fn is taken as input file. Note that the entire
  // command line will only be read if fn == "".
{
  std::string cfg_file_name,input_file=fn;
  bool dd=1;
  po::variables_map vm;
  po::options_description generic_options("GENERIC OPTIONS");
  generic_options.add_options()
    ("help,h","Produce help message")
    ("Copying,C","Return program version")
    ("config,c", po::value<std::string>(&cfg_file_name)->default_value(".spinner"), 
     "You can specify an alternative configuration file here.")
    ("input-file", po::value<std::string>(&input_file)->default_value("/dev/stdin"),
     "arg is a file describing the job details ")
    ;

  po::options_description lattice_options
    ("LATTICE TOPOLOGY DEFENITIONS (can also be specified in the config file)");
  lattice_options.add_options()
    ("lattice,L",po::value<std::string> (),
     "arg is a file describing the lattice.")
    ("La", po::value<uint>(), 
     "arg is number of unit cells along a-axis.")
    ("Lb", po::value<uint>(), 
     "arg is number of unit cells along b-axis.")
    ("Lc", po::value<uint>(), 
     "arg is number of unit cells along c-axis.") 
    ("lattice-params",po::value<std::string>(), 
     "Define lattice parameters a,b,c,alpha,beta,gamma.")
    ("add-atom",po::value<std::vector<std::string > >()->composing(),
     "Add atom to unit cell with position (x,y,z), spin quantum nr sn, g-factor gn and occupancy fn. Format is \"name xn yn zn sn gn fn\". The default is \"1 .0 .0 .0 1 2 1\".")
    ("add-bond",po::value<std::vector<std::string > >()->composing(),
     "Add exchange coupling to lattice (bond) with strength J (J > 0 = AF) between S_i and S_j a distance 'l' apart. Format is \" J l S_i S_j\". This option overwrites the legacy method of adding bonds, through the options J1 .. J4, which is retained for backwards compatibility.") 
    ("J1",po::value<double>()->default_value(1),
     "Nearest neighbour coupling constant.")
    ("J2",po::value<double>()->default_value(0),
     "Next nearest neighbour coupling constant.")
    ("J3",po::value<double>()->default_value(0),
     "3rd neighbour coupling constant.")   
    ("J4",po::value<double>()->default_value(0),
     "4th neighbour coupling constant.")   
    ("periodic-axes",po::value<std::string>(),
     "Axes along which periodic boundary conditions apply.")
    ("cut-off",po::value<double>(),
     "Maximum distance over which spins interact (with dipole hamiltonian).")
    ;

  po::options_description hamiltonian_options
    ("HAMILTONIAN DEFENITIONS (Can be specified in config ('.in') file and '.lattice' file)");
  hamiltonian_options.add_options()
    ("Hamiltonian",po::value<std::string>(),
     "Select either \'Potts\', \'super-exchange\', \'dipole-dipole reaction field\',\'dipole-dipole truncated\', \'dipole-dipole reaction field\', \'local MAE\' or \'DMI\'.")
    ("spin-dimension",po::value<int>()->default_value(3),
     "Select Ising(d=1), xy(d=2) or Heisenberg (d=3) spins. Note: When the dipole-dipole interaction is switched on, d=3 is forced.")
    ("Field,H",po::value<std::string>()->default_value("0 0 0"),
     "arg is a 3 dim vector defining an external applied field.")
    ("MAE",po::value<std::string>()->default_value("0 0 0"),
     "arg is a 3 dim vector defining the magnetic anisotropy energy.")
    //("Dz",po::value<double>()->default_value(0),
    // "arg is the strength of the Dzyaloshinskii-Moryia interaction.")
    ("DMI",po::value<std::string> (),
     "arg is a file describing the Dzyaloshinkii Moriya Interactions.")
    ("A_z",po::value<std::vector<std::string> >()->composing(),
     "Specify the anisotropy along the z-axis(for each ion in the unit cell).")
    ("A_xy",po::value<std::vector<std::string> >()->composing(),
     "Specify the anisotropy due to the crystal field in the xy plane for each ion in the unit cell).")
    ("a",po::value<std::vector<std::string> >()->composing(),
     "The cubic term of the local spin anisotropy, for each atom in unit cell.")
    ("phi",po::value<std::vector<std::string> >()->composing(),
     "phi, theta and psi specify the alignment of the local anisotropy for each ion in the unit cell.")
    ("theta",po::value<std::vector<std::string> >()->composing(),
     "phi, theta and psi specify the alignment of the local anisotropy for each ion in the unit cell.")
    ("psi",po::value<std::vector<std::string> >()->composing(),
     "phi, theta and psi specify the alignment of the local anisotropy for each ion in the unit cell.")
    ("kappa",po::value<double>(),
     "Spin distribution width in Ewald sum (required for Ewald sum hamiltonian). This is a double.")
    ("kpoints",po::value<std::string>(),
     "Number of points for sampling reciprocal space in Ewald sum (required with Ewald sum hamiltonian). This is a 3D vector.")
    ;

  po::options_description run_options
    ("RUN OPTIONS (can also be specified in the config ('.in') file)");
  run_options.add_options()
    ("T,T",po::value<std::vector<std::string> >()->composing(),
     "arg is list of temperatures")
    ("sweeps,s",po::value<uint>()->default_value(1000), 
     "arg is  number of required data points")
    ("export-state,E",
     "Set this keyword to save the final state of the system.")
    ("import-state,I",
     "Load the previously saved state.")
    ("no-averaging",
     "Set this keyword to save all datapoints, for hystogram analysis.")
    ("no-thermalise",
     "Set this keyword to start with all spins aligned along Z-axis, or in configuration as read from previous run with option '-R'")
    ;

  po::options_description dmi("DMI DEFENITION (specified in '.dmi' file, linked from option 'dmi' in hamiltonian defenition)");
  dmi.add_options()
    ("Dij", po::value<std::vector<std::string> >()->composing(),
     "Dzyaloshinkii Moriya vector for bond Si,Sj.")
    ("S1.site", po::value<std::vector<int> >() -> composing(),
     "required with Dij, specifies the site_id of S1 in Dij (Si x Sj)")
    ("S2.site",po::value<std::vector<int> >() -> composing(),
     "required with Dij, specifies the site_id of S2.")
    ("cell_offset",po::value<std::vector<std::string> >() -> composing(),
     "required with Dij, specifies relative index of S1.cell and S2.cell.")
    ;

  // parse the command line only if input-file comes from the command line.
  // if input-file (fn) is specified in function call, then only the 
  // contents of this file will be read into vm.
  if (fn=="") {
    po::positional_options_description p;
    p.add("input-file", 1);
    
    po::options_description cmdline_options;
    cmdline_options.add(generic_options).add(run_options).add(lattice_options).add(hamiltonian_options);
    
    po::store(po::command_line_parser(ac,av).options(cmdline_options).
	      positional(p).run(), vm);
    
    po::notify(vm);
  
    if (vm.count("help")) {
      std::cout << "\nNAME\n\tspinner  -  Monte-carlo simulation of ";
      std::cout << "the thermodynamics of \n\t\t  classical heisenberg ";
      std::cout << "spins on a user-defined lattice.\n\nSYNOPSYS\n\t";
      std::cout << "spinner 'input-file' [options ...]\n\n";
      std::cout << generic_options << "\n" << run_options << "\n"<< lattice_options
		<< "\n" << hamiltonian_options<<"\n"<<dmi<<"\n";
      exit(0);
    }    
    if (vm.count("Copying")) {
      show_copyright();
      std::cout <<"  Build: "<< __DATE__ << " " << __TIME__ << "\n\n";
      exit(0);
    }
  }
  else cfg_file_name = ".spinner";

  std::string base_name=input_file.substr(0,input_file.find('.',0));
  std::string output_file=base_name+".dat";
  std::string state_file=base_name+".state";
  std::string random_state='.'+base_name+".random_state";

  po::options_description 
    hidden_options("Hidden options, can only be specified in rc file");

  hidden_options.add_options()
    ("phi-max",po::value<double>()->default_value(3.1415),
     "    sets the maximum rotation of a newly generated spin around the current one.")
    ("si",po::value<uint>()->default_value(400),
     "    Calculate energy difference between datapoints 'si' sweeps apart.")
    ("sc",po::value<uint>()->default_value(10),
     "    Check stability criterium while equilibrating over sc datapoints.")
    ("sf",po::value<double>()->default_value(3.0),
     "    A higher value here will improve equilibration at the expense of time.")
    ("random-state", po::value<std::string>()->default_value(random_state),
     "    Saves the state of the random number generators for the next run.")
    ("output-file",po::value<std::string>()->default_value(output_file),"")
    ("state-file",po::value<std::string>()->default_value(state_file),"")
    ;

  //options are now defined. 
  //<<<<<<<<<<<<<<<<THIS IS WHERE THE PROGRAM STARTS>>>>>>>>>>>>>>>>
  //Now the options need to be read into the variables map 'vm'. 
  //Note there are four different sources; command line, input-file, lattice-file,
  //and a config file wich contains the hidden options.

  std::ifstream cfg_file,inp_file,latt_file,dmi_file;
  cfg_file.open(cfg_file_name.c_str(),std::ios::in);
  std::cout<<"\nReading "<<cfg_file_name.c_str()<<" ";
  if (!cfg_file) {
    cfg_file.clear();
    // stop here if a non-standard and non-existing config file was specified.
    if (cfg_file_name!=".spinner") throw po::invalid_option_value("Cannot find this file!");
    // otherwise, if .spinner is not in current directory, then look in home dir. 
    //`````````````the code commented out below is from an attempt to make it portable.
    //fs::path homedir(getenv("HOME"),fs::native);
    //homedir/=".spinner";
    //cfg_file_name=homedir.native_file_string();
    std::cout<<" - not found.\nReading $HOME/.spinner\n";
    cfg_file_name=std::string(getenv("HOME"))+"/.spinner";
    cfg_file.open(cfg_file_name.c_str(),std::ios::in);
    // if ~/.spinner does not exist, then create it. 
    if (!cfg_file) //(!fs::exists(cfg_file_name)) 
    {
      cfg_file.clear();
      std::cout<<" - not found.\nCreating "<<cfg_file_name<<std::flush;
      std::ofstream new_cfg;
      new_cfg.open(cfg_file_name.c_str(),std::ios::out);
      new_cfg<<"# This file is autogenerated by 'spinner'. The following\n";
      new_cfg<<"# options can be set here: phi-max, si, sc, sf, random-state\n";
      new_cfg<<"# output-file and state-file.\n";
      new_cfg.close();
      // now we can be sure that this file exists, so we open it. 
      cfg_file.open(cfg_file_name.c_str(),std::ios::in);
      if (!cfg_file) throw po::invalid_option_value(" - cannot create a config file in $HOME !");
      std::cout<<" ok\n";
    }
  }
  po::store(po::parse_config_file(cfg_file,hidden_options), vm);
  cfg_file.close();
  po::notify(vm);

  //by now, the input_file should always be there, either through 
  //function call or command line. 

  po::options_description config_options;
  config_options.add(generic_options).add(run_options).add(lattice_options).add(hamiltonian_options);

  po::options_description model_options;
  model_options.add(lattice_options).add(hamiltonian_options);

  std::cout<<"Reading "<<input_file<<'\n';
  inp_file.open(input_file.c_str(),std::ios::in);
  if (!inp_file) throw po::invalid_option_value("Cannot find this file!");
  po::store(po::parse_config_file(inp_file,config_options), vm);
  po::notify(vm);
  inp_file.close();

  if (vm.count("lattice") ) {
    std::cout<<"Reading "<<vm["lattice"].as<std::string>()<<'\n';
    latt_file.open(vm["lattice"].as<std::string>().c_str(),std::ios::in);
    if (!latt_file) throw po::invalid_option_value("Cannot find this file!");
    po::store(po::parse_config_file(latt_file,model_options), vm);
    po::notify(vm);
    latt_file.close();
  }

  if (vm.count("DMI") ) {
    std::cout<<"Reading "<<vm["DMI"].as<std::string>()<<'\n';
    dmi_file.open(vm["DMI"].as<std::string>().c_str(),std::ios::in);
    if (!dmi_file) throw po::invalid_option_value("Cannot find this file!");
    po::store(po::parse_config_file(dmi_file,dmi), vm);
    po::notify(vm);
    dmi_file.close();
  }

  //po::notify(vm);
  //all options are not read into vm.

  //Checking for missing parameters, and mutual consistency of selections.
  if (!vm.count("load-state")){
    if (!vm.count("La")|!vm.count("Lb")|!vm.count("Lc"))
      throw po::invalid_option_value("Missing entry 'La','Lb' or 'Lc' in input file.");
    
    if ((vm["La"].as<uint>()<=0)|(vm["Lb"].as<uint>()<=0)|(vm["Lc"].as<uint>()<=0)) 
      throw po::invalid_option_value("Lattice size must be at least 1 in each dimension.");
    
    if (!vm.count("periodic-axes"))
      throw po::invalid_option_value("Missing entry 'periodic-axes' in input-file.");
    
    if (!vm.count("lattice-params")) 
      throw po::invalid_option_value("Missing entry 'lattice-params' in input-file.");
  }

  if ((vm["spin-dimension"].as<int>()==1) and not ((vm["Hamiltonian"].as<std::string>()=="super-exchange") or 
						   (vm["Hamiltonian"].as<std::string>()=="Potts")))
    throw po::invalid_option_value("1D spin works only with \'Potts\' and \'super-exchange\' Hamiltonians");
  else if ((vm["spin-dimension"].as<int>()==2) and not (vm["Hamiltonian"].as<std::string>()=="super-exchange"))
    throw po::invalid_option_value("2D spin works only with \'super-exchange\' Hamiltonians");
  else if ((vm["Hamiltonian"].as<std::string>()=="dipole-dipole Ewald sum")  &  
      (!(vm.count("kappa") & vm.count("kpoints")) ) )
    throw po::invalid_option_value("'kappa' and 'kpoints' must both be defined with 'dipole_Ewald_sum' Hamiltonian.");
  
  return vm;
}  
