/* 
file : respin.cpp

Respin v.2, to analyse output from spinner, using the '.state' file.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "options.hpp"
#include "lattice_graph.hpp"
#include "Cfn.hpp"
#include <iostream>
#include <fstream>
  //#include "boost/graph/graphviz.hpp"

using namespace boost;

void print_energies(lattice_graph& lg)
{
   //energy calculation.
    double Ex,Ed,SiSj;
    Ex = 0; Ed = 0;
    edge_iterator edge,edges_end;
    for ( tie(edge,edges_end)=edges(lg);edge!=edges_end; edge++) {
      SiSj = ublas::inner_prod(lg[source(*edge,lg)].spin,
                               lg[target(*edge,lg)].spin);
      Ex += lg[*edge].J * SiSj;
      Ed += Ex + pow(lg[*edge].r_abs,-3) *
        lg[source(*edge,lg)].mu_eff * lg[target(*edge,lg)].mu_eff * 
        (SiSj - 3 * ublas::inner_prod(lg[source(*edge,lg)].spin,lg[*edge].ru)* 
                    ublas::inner_prod(lg[target(*edge,lg)].spin,lg[*edge].ru) );
    }
    std::cout<<"Total Energy w. super-exchange Hamiltonian = "<<Ex<<
      "\nTotal Energy w. dipole-dipole Hamiltonian = "<<Ed<<
      "\nNote that the the latter only makes sense for real systems."<<
      "\nAnd the sum of the two : "<<Ex+Ed;
}

std::map<int,std::deque<double> > MolecularFields(po::variables_map vm, lattice_graph& lg)
{
  euclid_vector Hmol(3),Hext(3);
  double Habs;
  Hext = parse_vector(3,vm["MolecularField"].as<std::string>());
  site_iterator site,first_site,last_site;
  o_edge_iterator oe;
  std::map<int,std::deque<double> > dm;
  tie(first_site,last_site)=vertices(lg);

  for (site=first_site;site!=last_site; site++) {
    Hmol = Hext; 
    
    for (oe = out_edges(*site,lg).first; oe!=out_edges(*site,lg).second; oe++){
      Hmol += lg[*oe].J * lg[target(*oe,lg)].spin; 
    }
    Habs = ublas::norm_2(Hmol);
    dm[lg[*site].site_index].push_back(Habs);
  }
  return dm;
}

po::variables_map parse_pros(int ac,char* av[])
{
  //only filenames and commandline options here... 
  po::variables_map vm;
  std::string vf("oob");
  po::options_description cmdline_options("COMMANDLINE OPTIONS");
  cmdline_options.add_options()
    ("help,h", "Produce help message")
    ("Copying,C", "Show copyright and warranty conditions")
    ("state-file", po::value<std::string>(),
     "this is the input argument for spinpros. arg is the .state file ")
    ("Energy,E","Prints energies")
    ("Vect,v","Converts the lattice graph from .state to a .vect file for use with geomview")
    ("Correlation,c","Calculates the pairwise correlation function, output to .CF file")
    ("Chirality,x", po::value<std::string>(),
     "Calculates the chiral structure. Has an Xcell file as argument, output to _X.state file")
    //("Graphviz,g","Create .dot file for visualisation of lattice graph with graphviz.")
    ("MolecularField,M",po::value<std::string>(),
     "Calculate the molecular fields for each site in unit cell.")
    //("Nematicity,N","Calculates the spin-nematic correlation function, output to .nem file")
    ;
  po::options_description further_options("Defenition of traingle over which to calculate chirality:");
  further_options.add_options()
    ("supercell", po::value<std::string>()->default_value("1 1 1"),
     "optional with -X, specifies the number of cells to be taken together along a,b and c axis/")
    ("S1.site", po::value<std::vector<int> >() -> composing(),
     "required with -X, specifies the site index of S1 in the calculation of the chirality of three spins")
    ("S2.site",po::value<std::vector<int> >() -> composing(),
     "required with -X, specifies the site index of S2.")
    ("S3.site",po::value<std::vector<int> >() -> composing(),
     "required with -X, specifies the site index of S3.")
    ("S1.cell",po::value<std::vector<std::string> >() -> composing(),
     "required with -X, specifies relative index of cell of S1.") 
    ("S2.cell",po::value<std::vector<std::string> >() -> composing(),
     "required with -X, specifies relative index of cell of S2.") 
    ("S3.cell",po::value<std::vector<std::string> >() -> composing(),
     "required with -X, specifies relative index of cell of S3.") 
    ;

  po::positional_options_description p;
  p.add("state-file", 1);
  po::store(po::command_line_parser(ac, av).options(cmdline_options).
	    positional(p).run(), vm);
  po::notify(vm);
  
  if (vm.count("help")) {
    std::cout << "\nNAME\n\trespin  -  For postprocessing of the \n"<<
      "of the output from spinner. \n\nSYNOPSYS\n\t"<< 
      "respin 'state-file' [options ...]\n\n"<<cmdline_options << "\n" <<
      further_options<<"\n";
    exit(0);
  }    
  if (vm.count("Copying")) {
    std::cout<<"\n\
  Respin v. 1.0\n\
\n\
  A Monte Carlo program for the simulation of 'classical'\n\
  magnetic systems.\n\
  Written by Mark de Vries, markadevries@gmail.com\n\
\n\
  Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of\n\
  Edinburgh, Old College, Edinburgh, UK\n\
\n\
  Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the\n\
  Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK\n\
\n\
  This program is free software; you can redistribute it and/or modify\n\
  it under the terms of the GNU General Public License as published by\n\
  the Free Software Foundation; either version 2 of the License, or\n\
  (at your option) any later version.\n\
\n\
  This program is distributed in the hope that it will be useful,\n\
  but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
  GNU General Public License for more details.\n\
\n\
  You should have received a copy of the GNU General Public License\n\
  along with this program; if not, write to the Free Software\n\
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA\n\
\n";
    std::cout << "  Build: "<< __DATE__ << " " << __TIME__ << "\n\n";
    exit(0);
  }

  if (vm.count("Chirality")) {
    std::string state_file = vm["state-file"].as<std::string>();
    std::string base_name = state_file.substr(0,state_file.find('.',0));
    std::string cfg_file_name = vm["Chirality"].as<std::string>();
    std::ifstream cfg_file;
    cfg_file.open(cfg_file_name.c_str(),std::ios::in);
    std::cout<<"\nReading "<<cfg_file_name.c_str()<<std::endl;
    if (!cfg_file) {
    cfg_file.clear();
    // stop here if a non-standard (and non-existing) config file was specified.
    if (cfg_file_name!=".spinner") throw po::invalid_option_value("Cannot find this file!");
    }
    po::store(po::parse_config_file(cfg_file,further_options), vm);
    cfg_file.close();
    po::notify(vm);
  }

  if (vm.count("Nematicity")) {
    if (vm.count("Chirality"))
      throw po::invalid_option_value("Chiral and Nematic order cannot be calculated in the same run because their options clash. Please do them one by one.");
    std::string state_file = vm["state-file"].as<std::string>();
    std::string base_name = state_file.substr(0,state_file.find('.',0));
    std::string cfg_file_name = base_name + ".Ncell";
    std::ifstream cfg_file;
    cfg_file.open(cfg_file_name.c_str(),std::ios::in);
    std::cout<<"\nReading "<<cfg_file_name.c_str()<<std::endl;
    if (!cfg_file) {
    cfg_file.clear();
    // stop here if a non-standard (or non-existing) config file was specified.
    if (cfg_file_name!=".spinner") throw po::invalid_option_value("Cannot find this file!");
    }
    po::store(po::parse_config_file(cfg_file,further_options), vm);
    cfg_file.close();
    po::notify(vm);
  }
  return vm;
}

int main(int ac ,char* av[] )
{ 
  
  try {

    po::variables_map pvm(parse_pros(ac,av));

    std::string state_file = pvm["state-file"].as<std::string>();
    std::string base_name = state_file.substr(0,state_file.find('.',0));
    std::string cfn_file=base_name+".CF";
    std::string xst_file=base_name+"_X.state";
    std::string nst_file=base_name+"_N.state";
    std::string input_file=base_name+".in";
    std::string vect_file=base_name+".vect";
    std::string dot_file=base_name+".dot";
    std::string MF_file=base_name+".MF";

    uint na,nb,nc;
    
    //std::string base_name=infn.substr(0,infn.find('.',0));
    std::fstream inf,of;
    inf.open(pvm["state-file"].as<std::string>().c_str(),std::ios::in);

    if (!inf)
      throw po::invalid_option_value("Cannot find state file.");
    std::cout<<"\nReading lattice_graph from "<<pvm["state-file"].as<std::string>()<<
      " ... "<<std::flush;
    lattice_graph lg(0);  
    inf>>lg;
    inf>>na>>nb>>nc;
    char s1[20], s2[90];
    inf.getline(s1,20); //this one reads the newline character after na nb nc. 
    inf.getline(s1,20); inf.getline(s2,90);
    std::string axes(s1); 
    std::string lattice_pars(s2);
    inf.close();
    std::cout<<"ok\n";
   
    if (pvm.count("Energy")) print_energies(lg);

    if (pvm.count("Vect")) {
      std::cout<<"Converting and saving lattice_graph to .vect format... "<<std::flush;
      of.open(vect_file.c_str(),std::ios::out);
      of<<out_VECT(lg);
      of.close();
      std::cout<<"ok\nSaving result to "<<vect_file<<std::endl;
    }

    if (pvm["Correlation"].as<double>() > 0 ) {
      std::cout<<"Calculating pairwise spin-correlation function... "<<std::flush;
      std::vector<std::map<double,double> > Cf;
      Cf=Cfn(lg,na,nb,nc,axes,lattice_pars);
      std::cout<<"ok\nSaving result to "<<cfn_file<<std::endl;
      of.open(cfn_file.c_str(),std::ios::out);
      std::map<double,double>::iterator Cf_it,Cf2_it=Cf[1].begin();
      of<<"#d\tCC\tCC2\n";
      for (Cf_it=Cf[0].begin();Cf_it!=Cf[0].end();Cf_it++) {
	of<<(*Cf_it).first<<"\t"<<(*Cf_it).second<<"\t"<<(*Cf2_it).second<<"\n";
	Cf2_it++;
      }
      of.close();
    }

    if (pvm.count("Chirality")) {
      std::cout<<"Calculating chirality lattice"<<std::flush;
      lattice_graph Xf(Chirality(pvm,lg));
      std::cout<<"ok\nSaving result to "<<xst_file<<"... "<<std::endl;
      of.open(xst_file.c_str(),std::ios::out);
      of<<Xf<<'\n'<<na<<' '<<nb<<' '<<nc<<'\n'<<axes<<'\n'<<lattice_pars<<'\n';
      of.close();
    }
    /*
    if (pvm.count("Nematicity")) {
      std::cout<<"Calculating spin-nematic correlation function... "<<std::flush;
      lattice_graph Nf(Nematicity(pvm,lg));
      std::cout<<"ok\nSaving result to "<<nst_file<<std::endl;
      of.open(nst_file.c_str(),std::ios::out);
      of<<Nf<<'\n'<<na<<' '<<nb<<' '<<nc<<'\n'<<axes<<'\n'<<lattice_pars<<'\n';
      of.close();
    }
    
    if (pvm.count("Graphviz")) {
      std::cout<<"Greating .dot file for visualisation of lattice graph with Graphviz"<<std::flush;
      std::cout<<"Saving result to "<<dot_file<<std::endl;
      of.open(dot_file.c_str(),std::ios::out);
      write_graphviz(of,lg);
      of.close();
    }
    */
    if(pvm.count("MolecularField")) {
      std::cout<<"Calculating molecular field for each site, sorted on site_index"<<std::endl;
      std::map<int,std::deque<double> > dm;
      std::map<int,std::deque<double> >::iterator m;
      dm = MolecularFields(pvm,lg);
      std::cout<<"Saving result to "<<MF_file<<std::endl;
      of.open(MF_file.c_str(),std::ios::out);
      of<<'#';
      std::cout<<"Output format: Each row contains the molecular fields\n for sites with identical site_index\n";
      std::cout<<"Length of each row: \n";
      for (m=dm.begin(); m!=dm.end(); m++) {
	of<< m->first <<'\t';
	std::cout<<m->second.size()<<'\t';
      }
      of<<'\n'; std::cout<<std::endl;
      for (m=dm.begin(); m!=dm.end(); m++) {
	while ( m->second.size() > 0 ) {
	  of<<m->second[0]<<'\t';
	  m->second.pop_front();
	}
	of<<'\n';
      }
      of.close();
    }
      
    std::cout<<std::endl;
  }
  
  catch(std::exception& er) 
    {
      std::cerr<<er.what() << "\n";
      return 1;
    }  

  return 1;

}
